import pytest

from weco import batch_query, build


# Internally, these functions use the WecoAI client
# therefore, we can test both the client and functional forms here
@pytest.fixture
def ml_task_evaluator():
    fn_name, version_number, _ = build(
        task_description="I want to evaluate the feasibility of a machine learning task. Give me a json object with three keys - 'feasibility', 'justification', and 'suggestions'.",
        multimodal=False
    )
    return fn_name, version_number


@pytest.fixture
def ml_task_inputs():
    return [
        {"text_input": "I want to train a model to predict house prices using the Boston Housing dataset hosted on Kaggle."},
        {"text_input": "I want to train a model to classify digits using the MNIST dataset hosted on Kaggle using a Google Colab notebook."},
    ]


@pytest.fixture
def image_evaluator():
    fn_name, version_number, _ = build(
        task_description="Describe the contents of the given images. Provide a json object with 'description' and 'objects' keys.",
        multimodal=True
    )
    return fn_name, version_number


@pytest.fixture
def image_inputs():
    return [
        {"images_input": ["https://www.integratedtreatmentservices.co.uk/wp-content/uploads/2013/12/Objects-of-Reference.jpg"]},
        {"images_input": ["https://t4.ftcdn.net/jpg/05/70/90/23/360_F_570902339_kNj1reH40GFXakTy98EmfiZHci2xvUCS.jpg"]},
    ]


def test_batch_query_text(ml_task_evaluator, ml_task_inputs):
    fn_name, version_number = ml_task_evaluator
    batch_inputs = ml_task_inputs

    query_responses = batch_query(fn_name=fn_name, version_number=version_number, batch_inputs=batch_inputs)

    assert len(query_responses) == len(batch_inputs)

    for query_response in query_responses:
        assert isinstance(query_response["output"], dict)
        assert isinstance(query_response["in_tokens"], int)
        assert isinstance(query_response["out_tokens"], int)
        assert isinstance(query_response["latency_ms"], float)

        output = query_response["output"]
        assert set(output.keys()) == {"feasibility", "justification", "suggestions"}


def test_batch_query_image(image_evaluator, image_inputs):
    fn_name, version_number = image_evaluator
    batch_inputs = image_inputs

    query_responses = batch_query(fn_name=fn_name, version_number=version_number, batch_inputs=batch_inputs)

    assert len(query_responses) == len(batch_inputs)

    for query_response in query_responses:
        assert isinstance(query_response["output"], dict)
        assert isinstance(query_response["in_tokens"], int)
        assert isinstance(query_response["out_tokens"], int)
        assert isinstance(query_response["latency_ms"], float)

        output = query_response["output"]
        assert set(output.keys()) == {"description", "objects"}


import pytest

from weco import abuild, aquery


# Internally, these functions use the WecoAI client
# therefore, we can test both the client and functional forms here
@pytest.mark.asyncio
async def test_abuild(text_evaluator, image_evaluator, text_and_image_evaluator):
    for evaluator in [text_evaluator, image_evaluator, text_and_image_evaluator]:
        fn_name, version_number, fn_desc = await evaluator
        assert isinstance(fn_name, str)
        assert isinstance(version_number, int)
        assert isinstance(fn_desc, str)


async def assert_query_response(query_response):
    assert isinstance(query_response, dict)
    assert isinstance(query_response["output"], dict)
    assert isinstance(query_response["in_tokens"], int)
    assert isinstance(query_response["out_tokens"], int)
    assert isinstance(query_response["latency_ms"], float)


@pytest.fixture
async def text_evaluator():
    fn_name, version_number, fn_desc = await abuild(
        task_description="Evaluate the sentiment of the given text. Provide a json object with 'sentiment' and 'explanation' keys.",
        multimodal=False
    )
    return fn_name, version_number, fn_desc


@pytest.mark.asyncio
async def test_text_aquery(text_evaluator):
    fn_name, version_number, _ = await text_evaluator
    query_response = await aquery(fn_name=fn_name, version_number=version_number, text_input="I love this product!")

    await assert_query_response(query_response)
    assert set(query_response["output"].keys()) == {"sentiment", "explanation"}


@pytest.fixture
async def image_evaluator():
    fn_name, version_number, fn_desc = await abuild(
        task_description="Describe the contents of the given images. Provide a json object with 'description' and 'objects' keys.",
        multimodal=True
    )
    return fn_name, version_number, fn_desc


@pytest.mark.asyncio
async def test_image_aquery(image_evaluator):
    fn_name, version_number, _ = await image_evaluator
    query_response = await aquery(
        fn_name=fn_name,
        version_number=version_number,
        images_input=[
            "https://www.integratedtreatmentservices.co.uk/wp-content/uploads/2013/12/Objects-of-Reference.jpg",
            "https://t4.ftcdn.net/jpg/05/70/90/23/360_F_570902339_kNj1reH40GFXakTy98EmfiZHci2xvUCS.jpg",
        ],
    )

    await assert_query_response(query_response)
    assert set(query_response["output"].keys()) == {"description", "objects"}


@pytest.fixture
async def text_and_image_evaluator():
    fn_name, version_number, fn_desc = await abuild(
        task_description="Evaluate, solve and arrive at a numerical answer for the image provided. Provide a json object with 'answer' and 'explanation' keys.",
        multimodal=True
    )
    return fn_name, version_number, fn_desc


@pytest.mark.asyncio
async def test_text_and_image_aquery(text_and_image_evaluator):
    fn_name, version_number, _ = await text_and_image_evaluator
    query_response = await aquery(
        fn_name=fn_name,
        version_number=version_number,
        text_input="Find x and y.",
        images_input=[
            "https://i.ytimg.com/vi/cblHUeq3bkE/hq720.jpg?sqp=-oaymwEhCK4FEIIDSFryq4qpAxMIARUAAAAAGAElAADIQj0AgKJD&rs=AOn4CLAKn3piY91QRCBzRgnzAPf7MPrjDQ"
        ],
    )

    await assert_query_response(query_response)
    assert set(query_response["output"].keys()) == {"answer", "explanation"}



import base64
import os
import random
import re
import string
from io import BytesIO
from typing import Dict, Optional, Tuple
from urllib.parse import urlparse

import requests
from PIL import Image


def is_local_image(maybe_local_image: str) -> bool:
    """
    Check if the file is a local image.

    Parameters
    ----------
    maybe_local_image : str
        The file path.

    Returns
    -------
    bool
        True if the file is a local image, False otherwise.
    """
    if not os.path.exists(maybe_local_image):  # Check if the file exists
        return False

    try:  # Check if the file is an image
        Image.open(maybe_local_image)
    except IOError:
        return False

    return True


def is_base64_image(maybe_base64: str) -> Tuple[bool, Optional[Dict[str, str]]]:
    """
    Check if the image is a base64 encoded image and if so, extract the image information from the encoded data or URL provided.

    Parameters
    ----------
    data : str
        The image data or URL.

    Returns
    -------
    Tuple[bool, Optional[Dict[str, str]]]
    """
    pattern = r"data:(?P<media_type>[\w/]+);(?P<source_type>\w+),(?P<encoding>.*)"
    match = re.match(pattern, maybe_base64)
    if match:
        return True, match.groupdict()

    return False, None


def is_public_url_image(maybe_url_image: str) -> bool:
    """
    Check if the string is a publicly accessible URL

    Parameters
    ----------
    maybe_url_image : str
        The URL to check.

    Returns
    -------
    bool
        True if the URL is publicly accessible, False otherwise.
    """
    try:
        # Check if it is a valid URL
        if not urlparse(maybe_url_image).scheme:
            return False

        # Check if the URL is publicly accessible
        response = requests.head(maybe_url_image)
        if response.status_code != 200:
            return False

        # Check if the URL is an image
        content_type = response.headers.get("content-type")
        if not content_type:
            return False
        if not content_type.startswith("image"):
            return False
    except Exception:
        return False

    return True


def get_image_size(image: str, source: str) -> float:
    """
    Get the size of the image in MB.

    Parameters
    ----------
    image : str
        The image data or URL.

    source : str
        The source of the image. It can be 'base64', 'url', or 'local'.

    Returns
    -------
    float
        The size of the image in MB.

    Raises
    ------
    ValueError
        If the image is not a valid image.
    """
    if source == "base64":
        _, base64_info = is_base64_image(maybe_base64=image)
        img_data = base64.b64decode(base64_info["encoding"])
    elif source == "url":
        response = requests.get(image)
        response.raise_for_status()
        img_data = response.content
    elif source == "local":
        with open(image, "rb") as f:
            img_data = f.read()
    else:
        raise ValueError("Invalid image input")

    img = Image.open(BytesIO(img_data))
    img_byte_arr = BytesIO()
    img.save(img_byte_arr, format=img.format)
    return img_byte_arr.tell() / 1000000  # MB


def preprocess_image(image: Image, file_type: str) -> Tuple:
    """
    Preprocess the image by converting it to RGB if it has an alpha channel.

    Parameters
    ----------
    image : Image
        The image to preprocess.
    file_type : str
        The file type of the image.

    Returns
    -------
    Image
        The preprocessed image.
    file_type : str
        The file type of the image.
    """
    # Do not rescale or resize. Only do this if latency becomes an issue.
    # Remove the alpha channel for PNG and WEBP images if it exists.
    if image.mode in ("RGBA", "LA") or (image.mode == "P" and "transparency" in image.info):
        image = image.convert("RGB")

    # If the image file type is JPG, convert to JPEG for PIL compatibility.
    if file_type == "jpg":
        file_type = "jpeg"
    return image, file_type


def generate_random_base16_code(length: int = 5):
    """
    Generate a random base16 code.

    Parameters
    ----------
    length : int
        The length of the code.

    Returns
    -------
    str
        The random base16 code.
    """
    return "".join(random.choices(string.hexdigits, k=length))



import asyncio
import base64
import os
import warnings
from io import BytesIO
from typing import Any, Callable, Coroutine, Dict, List, Optional, Tuple, Union

import httpx
import requests
from httpx import HTTPStatusError
from PIL import Image

from .constants import MAX_IMAGE_SIZE_MB, MAX_IMAGE_UPLOADS, MAX_TEXT_LENGTH, SUPPORTED_IMAGE_EXTENSIONS
from .utils import (
    generate_random_base16_code,
    get_image_size,
    is_base64_image,
    is_local_image,
    is_public_url_image,
    preprocess_image,
)


class WecoAI:
    """A client for the WecoAI function builder API that allows users to build and query specialized functions built by LLMs.
    The user must simply provide a task description to build a function, and then query the function with an input to get the result they need.
    Our client supports both synchronous and asynchronous request paradigms and uses HTTP/2 for faster communication with the API.

    Attributes
    ----------
    api_key : str
        The API key used for authentication.
    """

    def __init__(self, api_key: str = None, timeout: float = 120.0, http2: bool = True) -> None:
        """Initializes the WecoAI client with the provided API key and base URL.

        Parameters
        ----------
        api_key : str, optional
            The API key used for authentication. If not provided, the client will attempt to read it from the environment variable - WECO_API_KEY.

        timeout : float, optional
            The timeout for the HTTP requests in seconds (default is 30.0).

        http2 : bool, optional
            Whether to use HTTP/2 protocol for the HTTP requests (default is True).

        Raises
        ------
        ValueError
            If the API key is not provided to the client, is not set as an environment variable or is not a string.
        """
        # Manage the API key
        if api_key is None or not isinstance(api_key, str):
            try:
                api_key = os.environ["WECO_API_KEY"]
            except KeyError:
                raise ValueError("WECO_API_KEY must be passed to client or set as an environment variable")
        self.api_key = api_key
        self.http2 = http2
        self.timeout = timeout
        self.base_url = "https://function.api.weco.ai"
        # Setup clients
        self.client = httpx.Client(http2=http2, timeout=timeout)
        self.async_client = httpx.AsyncClient(http2=http2, timeout=timeout)

    def _headers(self) -> Dict[str, str]:
        """Constructs the headers for the API requests."""
        return {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

    def _make_request(self, endpoint: str, data: Dict[str, Any], is_async: bool = False) -> Callable:
        """Creates a callable for making either synchronous or asynchronous requests.

        Parameters
        ----------
        endpoint : str
            The API endpoint to which the request will be made.
        data : dict
            The data to be sent in the request body.
        is_async : bool, optional
            Whether to create an asynchronous request (default is False).

        Returns
        -------
        Callable
            A callable that performs the HTTP request.
        """
        url = f"{self.base_url}/{endpoint}"
        headers = self._headers()

        if is_async:

            async def _request():
                try:
                    response = await self.async_client.post(url, json=data, headers=headers)
                    response.raise_for_status()
                    return response.json()
                except HTTPStatusError as e:
                    # Handle HTTP errors (4xx and 5xx status codes)
                    error_message = f"HTTP error occurred: {e.response.status_code} - {e.response.text}"
                    raise ValueError(error_message) from e
                except Exception as e:
                    # Handle other exceptions
                    raise ValueError(f"An error occurred: {str(e)}") from e

            return _request()
        else:

            def _request():
                try:
                    response = self.client.post(url, json=data, headers=headers)
                    response.raise_for_status()
                    return response.json()
                except HTTPStatusError as e:
                    # Handle HTTP errors (4xx and 5xx status codes)
                    error_message = f"HTTP error occurred: {e.response.status_code} - {e.response.text}"
                    raise ValueError(error_message) from e
                except Exception as e:
                    # Handle other exceptions
                    raise ValueError(f"An error occurred: {str(e)}") from e

            return _request()

    def _process_query_response(self, response: Dict[str, Any]) -> Dict[str, Any]:
        """Processes the query response and handles warnings.

        Parameters
        ----------
        response : dict
            The raw API response.

        Returns
        -------
        dict
            A processed dictionary containing the output, token counts, and latency.

        Raises
        ------
        UserWarning
            If there are any warnings in the API response.
        """
        for _warning in response.get("warnings", []):
            warnings.warn(_warning)

        return {
            "output": response["response"],
            "in_tokens": response["num_input_tokens"],
            "out_tokens": response["num_output_tokens"],
            "latency_ms": response["latency_ms"],
        }

    def _build(self, task_description: str, multimodal: bool, is_async: bool) -> Union[Tuple[str, int, str], Coroutine[Any, Any, Tuple[str, int, str]]]:
        """Internal method to handle both synchronous and asynchronous build requests.

        Parameters
        ----------
        task_description : str
            A description of the task for which the function is being built.
        
        multimodal : bool
            Whether the function is multimodal or not.
        
        is_async : bool
            Whether to perform an asynchronous request.

        Returns
        -------
        Union[tuple[str, int, str], Coroutine[Any, Any, tuple[str, int, str]]]
            A tuple containing the name, version number and description of the function, or a coroutine that returns such a tuple.

        Raises
        ------
        ValueError
            If the task description is empty or exceeds the maximum length.
        """
        # Validate the input
        if len(task_description) == 0:
            raise ValueError("Task description must be provided.")
        if len(task_description) > MAX_TEXT_LENGTH:
            raise ValueError(f"Task description must be less than {MAX_TEXT_LENGTH} characters.")

        endpoint = "build"
        data = {"request": task_description, "multimodal": multimodal}
        request = self._make_request(endpoint=endpoint, data=data, is_async=is_async)

        if is_async:
            # return 0 for the version number
            async def _async_build():
                response = await request
                return response["function_name"], 0, response["description"]

            return _async_build()
        else:
            response = request  # the request has already been made and the response is available
            return response["function_name"], 0, response["description"]

    async def abuild(self, task_description: str, multimodal: bool = False) -> Tuple[str, int, str]:
        """Asynchronously builds a specialized function given a task description.

        Parameters
        ----------
        task_description : str
            A description of the task for which the function is being built.
        
        multimodal : bool, optional
            Whether the function is multimodal or not (default is False).

        Returns
        -------
        tuple[str, str]
            A tuple containing the name, version number and description of the function.
        """
        return await self._build(task_description=task_description, multimodal=multimodal, is_async=True)

    def build(self, task_description: str, multimodal: bool = False) -> Tuple[str, int, str]:
        """Synchronously builds a specialized function given a task description.

        Parameters
        ----------
        task_description : str
            A description of the task for which the function is being built.
        
        multimodal : bool, optional
            Whether the function is multimodal or not (default is False).

        Returns
        -------
        tuple[str, str]
            A tuple containing the name, version number and description of the function.
        """
        return self._build(task_description=task_description, multimodal=multimodal, is_async=False)

    def _upload_image(self, fn_name: str, version_number: int, upload_id: str, image_info: Dict[str, Any]) -> str:
        """
        Uploads an image to an S3 bucket and returns the URL of the uploaded image.

        Parameters
        ----------
        fn_name : str
            The name of the function for which the image is being uploaded.
        version_number : int
            The version number of the function for which the image is being uploaded.
        upload_id: str
            A unique identifier for the image upload.
        image_info : Dict[str, Any]
            A dictionary containing the image metadata.

        Returns
        -------
        str
            The URL of the uploaded image.
        """

        if image_info["source"] == "base64":
            _, base64_info = is_base64_image(maybe_base64=image_info["image"])
            img_data = base64.b64decode(base64_info["encoding"])
        elif image_info["source"] == "url":
            response = requests.get(image_info["image"])
            response.raise_for_status()
            img_data = response.content
        elif image_info["source"] == "local":
            with open(image_info["image"], "rb") as f:
                img_data = f.read()
        else:
            raise ValueError("Invalid image input")

        # Preprocess the image
        img = Image.open(BytesIO(img_data))
        file_type = image_info["file_type"]
        processed_img, file_type = preprocess_image(image=img, file_type=file_type)
        upload_data = BytesIO()
        processed_img.save(upload_data, format=file_type)
        upload_data = upload_data.getvalue()

        # Request a presigned URL from the server
        endpoint = "upload_link"
        request_data = {"fn_name": fn_name, "version_number": version_number, "upload_id": upload_id, "file_type": file_type}
        # This needs to be a synchronous request since we need the presigned URL to upload the image
        response = self._make_request(endpoint=endpoint, data=request_data, is_async=False)

        # Upload the image to the S3 bucket
        files = {"file": upload_data}
        http_response = requests.post(response["url"], data=response["fields"], files=files)
        if http_response.status_code == 204:
            pass
        else:
            raise ValueError("Image upload failed")

        # Return the URL of the uploaded image
        upload_link = f"{response['url']}{response['fields']['key']}"
        return upload_link

    def _validate_query(self, text_input: str, images_input: List[str]) -> List[Dict[str, Any]]:
        """
        Validate the input for the query method.

        Parameters
        ----------
        text_input : str
            The text input to the function.
        images_input : List[str]
            A list of image URLs or images encoded in base64 with their metadata to be sent as input to the function.

        Returns
        -------
        List[Dict[str, Any]]
            A list of dictionaries containing the image metadata.

        Raises
        ------
        ValueError
            If the input is invalid.
        """
        if not isinstance(text_input, str) or not isinstance(images_input, list):
            raise ValueError("Text input must be a string and images input must be a list of strings.")
        for image in images_input:
            if not isinstance(image, str):
                raise ValueError("Images input must be a list of strings.")

        # Assert that either text or images or both must be provded
        if len(text_input) == 0 and len(images_input) == 0:
            raise ValueError("Either text or images or both must be provided as input.")

        # Check if the text input is within the limit
        if len(text_input) > MAX_TEXT_LENGTH:
            raise ValueError(f"Text input must be less than {MAX_TEXT_LENGTH} characters.")

        # Check if the images input is within the limit
        if len(images_input) > MAX_IMAGE_UPLOADS:
            raise ValueError(f"Number of images must be less than {MAX_IMAGE_UPLOADS}.")

        # Check if input is an valid image
        image_info = []
        for image in images_input:
            is_base64, base64_info = is_base64_image(maybe_base64=image)
            if is_base64:
                file_type = base64_info["media_type"].split("/")[1]

            is_public_url = is_public_url_image(maybe_url_image=image)
            if is_public_url:
                response = requests.get(image)
                response.raise_for_status()
                file_type = response.headers["content-type"].split("/")[1]

            is_local = is_local_image(maybe_local_image=image)
            if is_local:
                file_type = os.path.splitext(image)[1][1:]

            if not (is_base64 or is_public_url or is_local):
                raise ValueError("Images must be local paths, public URLs or base64 encoded strings.")

            # Determine the source of image
            if is_base64:
                source = "base64"
            elif is_public_url:
                source = "url"
            elif is_local:
                source = "local"

            # Check if the image type is supported
            file_type = file_type.lower()
            if file_type not in SUPPORTED_IMAGE_EXTENSIONS:
                raise ValueError(
                    f"Image file type {file_type} is not supported. Supported types are {SUPPORTED_IMAGE_EXTENSIONS}."
                )

            # Check if the image size is within the limit
            size = get_image_size(image=image, source=source)
            if size > MAX_IMAGE_SIZE_MB:
                raise ValueError(f"Individual image sizes must be less than {MAX_IMAGE_SIZE_MB} MB each.")

            image_info.append({"image": image, "file_type": file_type, "size": size, "source": source})

        return image_info

    def _query(
        self, is_async: bool, fn_name: str, version_number: Optional[int], text_input: Optional[str], images_input: Optional[List[str]]
    ) -> Union[Dict[str, Any], Coroutine[Any, Any, Dict[str, Any]]]:
        """Internal method to handle both synchronous and asynchronous query requests.

        Parameters
        ----------
        is_async : bool
            Whether to perform an asynchronous request.
        fn_name : str
            The name of the function to query.
        version_number : int, optional
            The version number of the function to query.
        text_input : str, optional
            The text input to the function.
        images_input : List[str], optional
            A list of image URLs or images encoded in base64 with their metadata to be sent as input to the function.

        Returns
        -------
        Union[Dict[str, Any], Coroutine[Any, Any, dict]]
            A dictionary containing the query results, or a coroutine that returns such a dictionary.

        Raises
        ------
        ValueError
            If the input is invalid.
        """
        warnings.warn("Setting the version number of the function is not yet supported. Currently, the first version of the function will be used i.e., version 0.")
        version_number = 0
        # Validate the input
        image_info = self._validate_query(text_input=text_input, images_input=images_input)

        # Create links for all images that are not public URLs and upload images
        image_urls = []
        upload_id = generate_random_base16_code()
        for i, info in enumerate(image_info):
            if info["source"] == "url" or info["source"] == "base64" or info["source"] == "local":
                url = self._upload_image(fn_name=fn_name, version_number=version_number, upload_id=upload_id, image_info=info)
            else:
                raise ValueError(f"Image at index {i} must be a public URL or a path to a local image file.")
            image_urls.append(url)

        # Make the request
        endpoint = "query"
        data = {"name": fn_name, "text": text_input, "images": image_urls, "version_number": version_number}
        request = self._make_request(endpoint=endpoint, data=data, is_async=is_async)

        if is_async:

            async def _async_query():
                response = await request
                return self._process_query_response(response=response)

            return _async_query()
        else:
            response = request  # the request has already been made and the response is available
            return self._process_query_response(response=response)

    async def aquery(
        self, fn_name: str, version_number: Optional[int] = -1, text_input: Optional[str] = "", images_input: Optional[List[str]] = []
    ) -> Dict[str, Any]:
        """Asynchronously queries a function with the given function ID and input.

        Parameters
        ----------
        fn_name : str
            The name of the function to query.
        version_number : int, optional
            The version number of the function to query. If not provided, the latest version will be used. Pass -1 to use the latest version.
        text_input : str, optional
            The text input to the function.
        images_input : List[str], optional
            A list of image URLs or images encoded in base64 with their metadata to be sent as input to the function.

        Returns
        -------
        dict
            A dictionary containing the output of the function, the number of input tokens, the number of output tokens,
            and the latency in milliseconds.
        """
        return await self._query(fn_name=fn_name, version_number=version_number, text_input=text_input, images_input=images_input, is_async=True)

    def query(self, fn_name: str, version_number: Optional[int] = -1, text_input: Optional[str] = "", images_input: Optional[List[str]] = []) -> Dict[str, Any]:
        """Synchronously queries a function with the given function ID and input.

        Parameters
        ----------
        fn_name : str
            The name of the function to query.
        version_number : int, optional
            The version number of the function to query. If not provided, the latest version will be used. Pass -1 to use the latest version.
        text_input : str, optional
            The text input to the function.
        images_input : List[str], optional
            A list of image URLs or images encoded in base64 with their metadata to be sent as input to the function.

        Returns
        -------
        dict
               A dictionary containing the output of the function, the number of input tokens, the number of output tokens,
            and the latency in milliseconds.
        """
        return self._query(fn_name=fn_name, version_number=version_number, text_input=text_input, images_input=images_input, is_async=False)

    def batch_query(self, fn_name: str, batch_inputs: List[Dict[str, Any]], version_number: Optional[int] = -1) -> List[Dict[str, Any]]:
        """Batch queries a function version with a list of inputs.

        Parameters
        ----------
        fn_name : str
            The name of the function or a list of function names to query.

        batch_inputs : List[Dict[str, Any]]
            A list of inputs for the functions to query. The input must be a dictionary containing the data to be processed. e.g.,
            when providing for a text input, the dictionary should be {"text_input": "input text"}, for an image input, the dictionary should be {"images_input": ["url1", "url2", ...]}
            and for a combination of text and image inputs, the dictionary should be {"text_input": "input text", "images_input": ["url1", "url2", ...]}.
        
        version_number : int, optional
            The version number of the function to query. If not provided, the latest version will be used. Pass -1 to use the latest version.

        Returns
        -------
        List[Dict[str, Any]]
            A list of dictionaries, each containing the output of a function query,
            in the same order as the input queries.
        """
        async def run_queries():
            tasks = list(map(
                lambda fn_input: self.aquery(fn_name=fn_name, version_number=version_number, **fn_input),
                batch_inputs
            ))
            return await asyncio.gather(*tasks)

        return asyncio.run(run_queries())



from typing import Any, Dict, List, Optional

from .client import WecoAI


def build(task_description: str, multimodal: bool = False, api_key: str = None) -> tuple[str, int, str]:
    """Builds a specialized function synchronously given a task description.

    Parameters
    ----------
    task_description : str
        A description of the task for which the function is being built.
    multimodal : bool, optional
        A flag to indicate if the function should be multimodal. Default is False.
    api_key : str
        The API key for the WecoAI service. If not provided, the API key must be set using the environment variable - WECO_API_KEY.

    Returns
    -------
    tuple[str, str]
        A tuple containing the name and description of the function.
    """
    client = WecoAI(api_key=api_key)
    response = client.build(task_description=task_description, multimodal=multimodal)
    return response


async def abuild(task_description: str, multimodal: bool = False, api_key: str = None) -> tuple[str, int, str]:
    """Builds a specialized function asynchronously given a task description.

    Parameters
    ----------
    task_description : str
        A description of the task for which the function is being built.
    multimodal : bool, optional
        A flag to indicate if the function should be multimodal. Default is False.
    api_key : str
        The API key for the WecoAI service. If not provided, the API key must be set using the environment variable - WECO_API_KEY.

    Returns
    -------
    tuple[str, str]
        A tuple containing the name, version number and description of the function.
    """
    client = WecoAI(api_key=api_key)
    response = await client.abuild(task_description=task_description, multimodal=multimodal)
    return response


def query(
    fn_name: str, version_number: Optional[int] = -1, text_input: Optional[str] = "", images_input: Optional[List[str]] = [], api_key: Optional[str] = None
) -> Dict[str, Any]:
    """Queries a function synchronously with the given function ID and input.

    Parameters
    ----------
    fn_name : str
        The name of the function to query.
    version_number : int, optional
        The version number of the function to query. If not provided, the latest version is used. Default is -1 for the same behavior.
    text_input : str, optional
        The text input to the function.
    images_input : List[str], optional
        A list of image URLs or base64 encoded images to be used as input to the function.
    api_key : str
        The API key for the WecoAI service. If not provided, the API key must be set using the environment variable - WECO_API_KEY.

    Returns
    -------
    dict
        A dictionary containing the output of the function, the number of input tokens, the number of output tokens,
        and the latency in milliseconds.
    """
    client = WecoAI(api_key=api_key)
    response = client.query(fn_name=fn_name, version_number=version_number, text_input=text_input, images_input=images_input)
    return response


async def aquery(
    fn_name: str, version_number: Optional[int] = -1, text_input: Optional[str] = "", images_input: Optional[List[str]] = [], api_key: Optional[str] = None
) -> Dict[str, Any]:
    """Queries a function asynchronously with the given function ID and input.

    Parameters
    ----------
    fn_name : str
        The name of the function to query.
    version_number : int, optional
        The version number of the function to query. If not provided, the latest version is used. Default is -1 for the same behavior.
    text_input : str, optional
        The text input to the function.
    images_input : List[str], optional
        A list of image URLs to be used as input to the function.
    api_key : str
        The API key for the WecoAI service. If not provided, the API key must be set using the environment variable - WECO_API_KEY.

    Returns
    -------
    dict
        A dictionary containing the output of the function, the number of input tokens, the number of output tokens,
        and the latency in milliseconds.
    """
    client = WecoAI(api_key=api_key)
    response = await client.aquery(fn_name=fn_name, version_number=version_number, text_input=text_input, images_input=images_input)
    return response


def batch_query(
    fn_name: str, batch_inputs: List[Dict[str, Any]], version_number: Optional[int] = -1, api_key: Optional[str] = None
) -> List[Dict[str, Any]]:
    """Synchronously queries multiple functions using asynchronous calls internally.

    This method uses the asynchronous queries to submit all queries concurrently
    and waits for all responses to be received before returning the results.

    Parameters
    ----------
    fn_name : str | List[str]
        The name of the function or a list of function names to query.
        Note that if a single function name is provided, it will be used for all queries.
        If a list of function names is provided, the length must match the number of queries.

    batch_inputs : List[str]
       A list of inputs for the functions to query. The input must be a dictionary containing the data to be processed. e.g.,
       when providing for a text input, the dictionary should be {"text_input": "input text"}, for an image input, the dictionary should be {"images_input": ["url1", "url2", ...]}
       and for a combination of text and image inputs, the dictionary should be {"text_input": "input text", "images_input": ["url1", "url2", ...]}.

    version_number : int, optional
        The version number of the function to query. If not provided, the latest version is used. Default is -1 for the same behavior.

    api_key : str, optional
        The API key for the WecoAI service. If not provided, the API key must be set using the environment variable - WECO_API_KEY.

    Returns
    -------
    List[Dict[str, Any]]
        A list of dictionaries, each containing the output of a function query,
        in the same order as the input queries.
    """
    client = WecoAI(api_key=api_key)
    responses = client.batch_query(fn_name=fn_name, version_number=version_number, batch_inputs=batch_inputs)
    return responses
